import re
from argparse import ArgumentParser
import argparse
from jlp.launchpadutils import build_state, LaunchpadVote
from jlp import (launchpadutils, MergeProposalReview, Branch,
                 DputRunner, jenkinsutils, get_launchpad,
                 get_config_option, logger)
from tarmac.exceptions import TarmacMergeError, BranchHasConflicts
from bzrlib.errors import LockFailed
import os
from shutil import rmtree


class StringSplitAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.split(','))


class StringStripAction(argparse.Action):
    """Strips double quotes added by Jenkins for variables ending with '}'"""
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.strip('"'))


def get_fixes_commit_message(mp):
    ret = ''
    for b in mp.source_branch.linked_bugs_collection:
        ret = ret + ', ' + b.web_link
    ret = ret.replace(', ', '', 1)
    if ret == '':
        return ''
    else:
        return ' Fixes: ' + ret + '.'


def get_approved_by_commit_message(mp):
    ret = ''
    for vote in mp.votes:
        if vote.comment and vote.comment.vote == LaunchpadVote.APPROVE:
            ret = ret + ', ' + vote.reviewer.display_name
    ret = ret.replace(', ', '', 1)
    return ret


def format_commit_message(mp, commit_message):
    formatted_message = u"{commit_message}.{fixes}"
    #get rid of the trailing '.' so we don't end up with '..' when
    #people use '.' at the end of their commit message
    commit_message = commit_message.rstrip('.')
    formatted_message = formatted_message.format(
        commit_message=commit_message,
        fixes=get_fixes_commit_message(mp))
    approved_by = get_approved_by_commit_message(mp)
    if mp and not approved_by:
        approved_by = mp.reviewer.display_name
    if approved_by:
        formatted_message += u"\n\nApproved by {approved_by}.".format(
            approved_by=approved_by)
    return formatted_message


def change_mp_status(mp, reason, args, vote):
    args[reason] = True
    mp_state, message, subject = build_state(args)
    message = jenkinsutils.format_message_for_mp_update(
        args['build_job_url'],
        message,
        include_rebuild_link=False)
    launchpadutils.change_mp_status(mp, mp_state, message, subject,
                                    args['revision'], vote)


def merge_and_commit(mp, args):
    #if there is no reviewed_revid then something went wrong
    if not mp.reviewed_revid:
        logger.debug('Approved revid is not set '
                     '(maybe a permission problem?). Failing autolanding.')
        change_mp_status(mp, 'invalid_revid',
                         args, LaunchpadVote.NEEDS_FIXING)
        return 1

    target = Branch.create(mp.target_branch, config=False, create_tree=True)
    source = Branch.create(mp.source_branch, config=False, target=target)
    revid_str = str(mp.reviewed_revid)
    approved = source.bzr_branch.revision_id_to_revno(revid_str)

    logger.debug('mp.reviewed_revid is %s' % mp.reviewed_revid)
    #if unapproved changes made after approval do not land
    if source.bzr_branch.revno() > approved:
        logger.debug('Unapproved changes made after approval. '
                     'Failing autolanding.')
        change_mp_status(mp, 'unapproved_changes',
                         args, LaunchpadVote.NEEDS_FIXING)
        return 1

    try:
        logger.debug('Merging %(source)s to %(target)s' % {
                     'source': str(mp.source_branch.bzr_identity),
                     'target': str(mp.target_branch.bzr_identity)})
        target.merge(source, str(mp.reviewed_revid))
        reviews = launchpadutils.get_reviews(mp)
        logger.debug('Commiting previous merge')
        commit_message = launchpadutils.get_commit_message(
            mp,
            args['use_description_for_commit'])
        msg = format_commit_message(mp, commit_message)
        target.commit(msg,
                      revprops={'merge_url': mp.web_link},
                      authors=source.authors,
                      reviews=reviews)
        revision = target.bzr_branch.revno()
        logger.debug('New revision is: ' + str(revision))
        logger.debug('Closing bugs')
        launchpadutils.close_bugs(mp)
        dputResult = True
        if args['ppa']:
            # try to extract the build job number from the build_job_url
            #its usually the last number in the URL
            jenkins_build_number = 0
            if args['build_job_url']:
                m = re.match('.*/([0-9]+)/?$', args['build_job_url'])
                if m:
                    jenkins_build_number = m.group(1)

            dputRunner = DputRunner(
                ppas=args['ppa'],
                trunk=target.branch_url,
                packagingBranch=args['packaging_branch'],
                revision=revision,
                DEBEMAIL=get_config_option('DEBEMAIL'),
                DEBFULLNAME=get_config_option('DEBFULLNAME'),
                distributions=args['ppa_distributions'],
                versionFormat=args['version_string_format'],
                jenkinsBuildNumber=jenkins_build_number,
                hooks=args['hooks'],
                release_only=args['release_only'])
            dputResult = dputRunner.dput()
        if not dputResult:
            logger.debug('Dput failed')
            return 1
    except BranchHasConflicts:
        logger.info('Merging failed - branch has conflicts')
        change_mp_status(mp, 'merge_failed', args,
                         LaunchpadVote.NEEDS_FIXING)
        return 1
    except TarmacMergeError:
        logger.info('Landing failed')
        change_mp_status(mp, 'landing_failed', args,
                         LaunchpadVote.NEEDS_FIXING)
        return 1
    except LockFailed:
        logger.info('Bazaar error: LockFailed. Jenkins user is maybe not ' +
                    'allowed to push into target?')
        change_mp_status(mp, 'landing_failed', args,
                         LaunchpadVote.NEEDS_FIXING)
        return 1
    finally:
        target.cleanup()
        review = MergeProposalReview(mp)
        review.unlock()

    return 0


def get_last_mp_vote(mp):
    for vote in mp.votes:
        launchpad_login = get_config_option('launchpad_login')
        if vote.registrant.name == launchpad_login and vote.comment:
            return vote.comment.vote
    return None


def autoland():
    parser = ArgumentParser(description="Start autolanding process " +
                            "resulting from a merge proposal.")
    parser.add_argument('-p', '--ppa',
                        help="comma separated list of PPAs where the " +
                             "merged result shuld be dput-ed",
                        action=StringSplitAction)
    parser.add_argument('-a', '--packaging-branch',
                        help="Packaging branch to be used for dput-ing")
    parser.add_argument('-d', '--ppa-distributions', default=None,
                        help="Comma separated list of distributions " +
                        "(e.g. precise,quantal)",
                        action=StringSplitAction)
    parser.add_argument('-o', '--hooks', default=[],
                        help="Comma separated list of hack hooks" +
                        "that should be used before dput " +
                        "(e.g. H40native_hack.py,H10autoreconf)",
                        action=StringSplitAction)
    parser.add_argument('-r', '--test-result', required=True,
                        help="Result of previous testing" +
                        "Autolanding will be stopped if this is FAILED")
    parser.add_argument('-v', '--revision', required=True,
                        help="Revision of source branch that should be merged")
    parser.add_argument('-s', '--source-branch',
                        help='Not used but kept for backwards compatibility.')
    parser.add_argument('-m', '--merge-proposal', required=True,
                        help="Merge proposal which is supposed to be merged")
    parser.add_argument('-b', '--build-job-url',
                        help="URL to the build/test job which was " +
                             "previously executed")
    parser.add_argument('-f', '--version-string-format',
                        help="Format string for the version which will be " +
                              "used to dput the package " +
                              "(only used when --ppa specified)",
                        action=StringStripAction)
    parser.add_argument('-u', '--use-description-for-commit',
                        default=False,
                        action='store_true',
                        help="Use the description field of merge proposal " +
                             "if no commit message is specified")
    parser.add_argument('-l', '--release-only',
                        default=False,
                        action='store_true',
                        help="Only do dch --release when dputing into PPA " +
                             "and don't modify the changlog in any other way")

    args = vars(parser.parse_args())

    # launchpadlib is not thread/process safe so we are creating launchpadlib
    # cache in /tmp per process which gets cleaned up at the end
    # see also lp:459418 and lp:1025153
    launchpad_cachedir = os.path.join('/tmp',
                                      str(os.getpid()),
                                      '.launchpadlib')
    lp_handle = get_launchpad(launchpadlib_dir=launchpad_cachedir)
    mp = launchpadutils.get_mp_handle_from_url(lp_handle,
                                               args['merge_proposal'])

    #check for test_result and fail if previous testing failed
    if args['test_result'] == 'FAILED':
        logger.debug('Previous test failed. Failing autolanding.')
        change_mp_status(mp, 'landing_failed', args,
                         LaunchpadVote.NEEDS_FIXING)
        return 0

    #check for commit message and fail if it is empty
    #if --use-description-for-commit is specified fallback for description
    # and if description is not set then fail
    if not (launchpadutils.get_commit_message(
            mp, args['use_description_for_commit'])):
        logger.debug('Commit message empty. Failing autolanding.')
        change_mp_status(mp, 'commit_message_empty',
                         args, LaunchpadVote.NEEDS_FIXING)
        return 1

    # Approve the merge proposal in case the last jenkins vote was
    # different from Approve (this includes no vote)
    if get_last_mp_vote(mp) != LaunchpadVote.APPROVE:
        mp.createComment(
            review_type=get_config_option('launchpad_review_type'),
            vote=LaunchpadVote.APPROVE,
            subject='Re: [Merge] autolanding tests passed')

    # merging starts here
    ret = merge_and_commit(mp, args)
    rmtree(launchpad_cachedir, ignore_errors=True)
    return ret
