import re
from jlp import logger, get_config_option
from .mergeproposalreview import MergeProposalReview
import jenkinsutils
from . import get_json_jenkins


class LaunchpadVote():
    APPROVE = 'Approve'
    DISAPPROVE = 'Disapprove'
    NEEDS_FIXING = 'Needs Fixing'


def get_branch_from_mp(merge_proposal):
    """Return a link to branch given a link to a merge proposal.

    If merge_proposal is:
      https://copde.launchpad.net/~user/project/name/+merge/12345
    then the result will be:
      https://copde.launchpad.net/~user/project/name/

    :param merge_proposal: url of a launchpad merge proposal
    """
    m = re.search('(.*)\+merge/[0-9]+$', merge_proposal)
    if m:
        return m.group(1)
    return None


def change_mp_status(mp, status, message, subject, revision, vote=None):
    """Change status (vote) of merge proposal.

    :param status: merge proposal status (e.g. "Needs review")
    :param message: message for the vote
    :param subject: subject for the e-mail with the message
    :param revision: revision against which this message is valid
    :param vote: None or something from LaunchpadVote
                 (e.g. LaunchpadVote.NEEDS_FIXING)
    """
    mp.setStatus(status=status, revid=revision)
    if vote:
        mp.createComment(
            review_type=get_config_option('launchpad_review_type'),
            subject=subject, content=message, vote=vote)
        review = MergeProposalReview(mp)
        review.unlock()
    else:
        mp.createComment(
            review_type=get_config_option('launchpad_review_type'),
            subject=subject, content=message)


def close_bugs(mp):
    """ Close bugs that are linked to this merge proposal.

    This function goes through all linked bug reports and closes all bugs
    """
    for b in mp.source_branch.linked_bugs_collection:
        for task in b.bug_tasks_collection:
            if task.target.web_link == mp.target_branch.project.web_link:
                task.status = 'Fix Committed'
                task.lp_save()
                message = 'Fix committed into {target} at revision ' + \
                          '{revision}, scheduled for release in ' + \
                          '{project}, milestone {milestone}'
                mp.lp_refresh()
                project = mp.target_branch.project
                milestone_name = 'Unknown'
                active_milestones = project.active_milestones_collection
                for active_milestone in active_milestones:
                    series = active_milestone.series_target
                    if series.branch_link == mp.target_branch.self_link:
                        milestone_name = active_milestone.name
                message = message.format(
                    target=mp.target_branch.display_name,
                    revision=mp.merged_revno,
                    project=project.name,
                    milestone=milestone_name)
                b.newMessage(content=message)


def get_reviews(mp):
    """For a given merge proposal get all reviews.

    Returns list of reviews or None if there are no reviews so far.

    :param mp: merge proposal
    """

    reviews = []
    for vote in mp.votes:
        if vote.comment:
            reviews.append('%s;%s' % (vote.reviewer.display_name,
                                      vote.comment.vote))
    if len(reviews) == 0:
        return None

    return reviews


def get_commit_message(mp, use_description):
    """Get commit message for a given merge proposal.

    If use_description is True and there is no commit message a descritpion
    will be returned instead.

    :param use_description: boolean flag indicating whether we should fallback
                            to description for commit message
    """
    message = None
    if mp.commit_message:
        message = mp.commit_message
    elif use_description and mp.description:
        message = mp.description
    return message


def is_commit_message_set(mp, jenkins_job, jenkins_url):
    """Check if commit message is defined for a given merge proposal.

    This also checks the given jenkins job for an use_description_for_commit
    parameter. If it exists in the jenkins job then the commit message
    fallbacks to description.

    :param mp: launchpad handle to merge proposal
    :param jenkins_job: name of the jenkins job (use_description_for_commit is
                        searched here)
    :param jenkins_url: jenkins url
    """

    jenkins = get_json_jenkins()
    job_url = jenkins_url + '/job/' + jenkins_job + '/'
    use_description = False
    try:
        if jenkinsutils.get_parameter(jenkins, job_url,
                                      'use_description_for_commit'):
            use_description = True
    except KeyError:
        logger.debug("Unable to get \"use_description_for_commit\" " +
                     "configuration for %s" % (jenkins_job))
    logger.debug('Job is configured to fallback to description: %s' %
                 (use_description))
    return bool(get_commit_message(mp, use_description))


def unapproved_prerequisite_exists(mp):
    """Check if there is an unapproved prerequisite branch for a given merge
    proposal.

    :param mp: launchpadlib handle to merge proposal
    """
    prereq = mp.prerequisite_branch
    target_branch = mp.target_branch
    if prereq:
        #x is a merge-proposal object
        merges = [x for x in prereq.landing_targets
                  if x.target_branch.web_link == target_branch.web_link and
                  x.queue_status != u'Superseded']
        if len(merges) == 0:
            logger.debug('No proposals of prerequisite branch exists')
            return True
        else:
            for merge in merges:
                #lets make our life simpler here and lets assume
                #all the MP needs to be merged even though in reality
                #their target branch might differ
                if merge.queue_status != u'Merged':
                    logger.debug('Prerequisite not yet merged')
                    return True
    #TODO
    # skip also branches with changes made after approval
    return False


def testing_in_progress(mp):
    """ Returns whether testing is in progress.

    We are using a file (represented by MergeProposalReview class) to indicate
    a testing is in progress. This file is created when the jenkins job is
    started and cleaned up either by generic-update-mp job or generic-land.

    :param mp: launchpad merge proposal handle
    """
    review = MergeProposalReview(mp)
    if review.is_locked():
        logger.debug('Skipping this MP. It is curently being ' +
                     'tested by Jenkins.')
        return True
    return False


def user_allowed_to_trigger_jobs(lp_user):
    """Returns if a given user is allowed to run jobs on jenkins.

    This is to avoid random people to run jobs on our internal infrastructure.
    User is allowed if they are either directly in the ALLOWED_USERS list or
    are member of a team in that list.

    :param lp_user: launchpad user handle
    """
    allowed_people = get_config_option('allowed_users')
    if lp_user.name in allowed_people:
        return True
    for membership in lp_user.memberships_details:
        if membership.team.name in allowed_people:
            return True
    logger.debug('User "' + lp_user.name +
                 '" not allowed to trigger jobs')
    return False


def is_merge_only(mp):
    """Determines if an mp can be merged without a build"""
    # diffstat is a dictionary of {file: changeset}
    # diffstat will be None if launchpad has not yet generated a diff
    if mp.preview_diff is None or mp.preview_diff.diffstat is None:
        raise LookupError("preview_diff of this merge proposal is None")
    diff_stat = mp.preview_diff.diffstat
    logger.debug("%s" % (diff_stat))
    if (len(diff_stat) == 1 and u'debian/changelog' in diff_stat):
        logger.debug("debian/changelog only merge")
        return True
    return False


def get_latest_review(launchpad_user, mp):
    """Return the latest revision reviewed by the given launchpad_user.

    This function expects a review comment in the following format:
       '^(PASSED|FAILED): Continuous integration, rev:(\d+)'

    :param launchpad_user: handle to launchpad user
    :param mp: handle to merge proposal
    """
    revision = 0
    launchpad_review_type = get_config_option('launchpad_review_type')
    for comment in mp.all_comments:
        if comment.author.name == launchpad_user.name:
            if comment.vote_tag == launchpad_review_type:
                m = re.search(
                    '^(PASSED|FAILED): Continuous integration, rev:(\d+)',
                    comment.message_body)
                if m and (int(m.group(2)) > revision):
                    revision = int(m.group(2))
    return revision


def latest_candidate_validated(launchpad_user, mp):
    """Return if the latest candidate revision of the merge proposal is
    validated.

    :param launchpad_user: handle to launchpad user used to validate the
                           merge proposals
    :param mp: handle to merge proposal
    """

    latest_review = get_latest_review(launchpad_user, mp)
    logger.debug('Latest_review is revision: ' + str(latest_review))
    if latest_review >= mp.source_branch.revision_count:
        logger.debug('Skipping this MP. Current revision: ' +
                     str(mp.source_branch.revision_count))
        return True
    return False


def approve_mp(mp, revision, build_url):
    """Approve a given merge proposal a revision.

    :params mp: launchapd handle to the respective merge proposal
    :params revision: revision that should be approved
    :params build_url: jenkins build url with the details. This job is used to
                       generate the message with all the links to test runs as
                       well as artifacts (coverity, deb files, etc)
    """
    state = 'PASSED: Continuous integration, rev:' + str(revision)
    logger.debug(state)
    content = jenkinsutils.format_message_for_mp_update(build_url,
                                                        state + "\n")
    mp.createComment(review_type=get_config_option('launchpad_review_type'),
                     vote=LaunchpadVote.APPROVE, subject=state,
                     content=content)
    review = MergeProposalReview(mp)
    review.unlock()


def disapprove_mp(mp, revision, build_url, reason=None):
    """Disapprove a given merge proposal a revision (vote Needs Fixing).

    :params mp: launchapd handle to the respective merge proposal
    :params revision: revision that should be fixed
    :params build_url: jenkins build url with the details. This job is used to
                       generate the message with all the links to test runs as
                       well as artifacts (coverity, deb files, etc)
    :params reason: optional string that is attached to the comment
    """
    state = "FAILED: Continuous integration, rev:{revision}".format(
        revision=revision)
    subject = state
    if reason:
        state = "{state}\n{reason}".format(state=state, reason=reason)

    logger.debug(state)
    content = jenkinsutils.format_message_for_mp_update(
        build_url, state + "\n")
    mp.createComment(review_type=get_config_option('launchpad_review_type'),
                     vote=LaunchpadVote.NEEDS_FIXING, subject=subject,
                     content=content)
    review = MergeProposalReview(mp)
    review.unlock()


class UpdateMergeProposalException(Exception):
    pass


def build_state(args):
    """Return a tuple containing merge proposal state, message, and
    subject."""

    if 'merge_failed' in args and args['merge_failed']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Merging failed. More details in the following jenkins job:\n",
                "Re: [Merge] autolanding failed (merge failed)")
    elif 'push_failed' in args and args['push_failed']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Pushing failed. More details in the following jenkins job:\n",
                "Re: [Merge] autolanding failed (push failed)")
    elif 'landing_failed' in args and args['landing_failed']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "More details in the following jenkins job:\n",
                "Re: [Merge] autolanding failed")
    elif 'merged' in args and args['merged']:
        return ('Merged',
                "PASSED: Autolanding.\nBranch merged.",
                "Re: [Merge] branch merged")
    elif 'commit_message_empty' in args and args['commit_message_empty']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "No commit message was specified in the merge proposal. " +
                "Hit 'Add commit message' on the merge proposal web page " +
                "or follow the link below. You can approve the merge " +
                "proposal yourself to rerun.\n" +
                args['merge_proposal'] + "/+edit-commit-message\n",
                "Re: [Merge] autolanding failed",)
    elif 'unapproved_changes' in args and args['unapproved_changes']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Unapproved changes made after approval.\n",
                "Re: [Merge] autolanding failed",)
    elif 'invalid_revid' in args and args['invalid_revid']:
        return ('Needs review',
                "FAILED: Autolanding.\n" +
                "Approved revid is not set in launchpad " +
                "(maybe a permission problem?).\n",
                "Re: [Merge] autolanding failed",)
    else:
        raise UpdateMergeProposalException("unrecognized parameters")


def get_merge_proposals(lp_handle, branch_name, state):
    """ Return all merge proposal for a given branch.

    Returns a list of launchpad handles for merge proposals.
    :param lp_handle: launchpad API handle/instance
    :param branch_name: name of the branch
    :param state: list of states I'm interested in (usually something like
    ['Approved'] or ['Needs review'])
    """
    branch = lp_handle.branches.getByUrl(url=branch_name)
    if not branch:
        logger.debug('Branch "' + branch_name + '" doesn\'t exist')
        return []
    mps = []
    for mp in branch.landing_candidates:
        if mp.queue_status in state:
            mps.append(mp)
    return mps


def get_mp_handle_from_url(lp_handle, merge_proposal_link):
    """ Get launchpad handle for merge proposal given a merge proposal URL.

    Returns None in case the merge proposal can't be found.
    :param merge_proposal_link: URL of the merge proposal
    """
    branch_link = get_branch_from_mp(merge_proposal_link)
    if not branch_link:
        logger.error('Unable to get branch link from merge proposal link.')
        return None

    branch_link = branch_link.replace('https://code.launchpad.net/', 'lp:')
    branch_link = branch_link.replace('https://code.staging.launchpad.net/',
                                      'lp://staging/')
    logger.debug('fetching branch: ' + branch_link)
    branch = lp_handle.branches.getByUrl(url=branch_link)
    if not branch:
        logger.debug('Branch {} does not exist'.format(branch_link))
        return None

    logger.debug('mp_link: {}.'.format(merge_proposal_link))

    for mp in branch.landing_targets:
        logger.debug('mp.web_link: {}'.format(mp.web_link))
        if mp.web_link == merge_proposal_link:
            return mp

    return None
