from mock import MagicMock, patch
from jlp import jenkinsutils, launchpadutils
import unittest
from tests import JenkinsJSONData
from jlp import get_config_option
from testscenarios import TestWithScenarios
from jenkins import JenkinsException
from lazr.restfulclient.errors import Unauthorized
from textwrap import dedent


class TestNormalizeURL(TestWithScenarios, unittest.TestCase):
    scenarios = [
        ('double-slashes',
         {'url': 'http://asd//ds',
          'expected': 'http://asd/ds'}),
        ('several-double-slashes',
         {'url': 'http://1//2//3//4',
          'expected': 'http://1/2/3/4'}),
    ]

    def test_normalize_url(self):
        self.assertEqual(jenkinsutils.normalize_url(self.url), self.expected)


class JenkinsUtilsGetDownstreamProjects(unittest.TestCase):
    def setUp(self):
        self.jenkins = MagicMock()
        self.jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects': [{'name': 'project1'},
                                                 {'name': 'project2'},
                                                 {'name': 'project3'}]})

    def test_get_downstream_projects(self):
        ret = jenkinsutils.get_downstream_projects(self.jenkins,
                                                   'http://ip/job/x/55')
        expected = ['project1', 'project2', 'project3']
        self.assertEqual(ret, expected)

    def test_get_downstream_projects_none(self):
        self.jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects': []})
        ret = jenkinsutils.get_downstream_projects(self.jenkins,
                                                   'http://ip/job/x')
        self.assertEqual(ret, [])

    def test_get_merge_job(self):
        self.jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects':
                          [{"name": "generic-cleanup"},
                           {"name": "generic-land"}]})
        ret = jenkinsutils.get_merge_job(self.jenkins, 'http://ip/job/x')
        self.assertEqual(ret, 'generic-land')

    def test_get_merge_job_none(self):
        self.jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects': []})
        with self.assertRaises(KeyError) as err:
            jenkinsutils.get_merge_job(self.jenkins, 'http://ip/job/x')
        self.assertIn("No merge jobs found", "%s" % (err.exception))

    def test_get_merge_job_too_many(self):
        self.jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects':
                          [{'name': 'my-land'},
                          {'name': 'your-land'}]})
        with self.assertRaises(KeyError) as err:
            jenkinsutils.get_merge_job(self.jenkins, 'http://ip/job/x')
        self.assertIn("Multiple merge jobs found: my-land, your-land",
                      "%s" % (err.exception))


class JenkinsUtilsGetDownstreamBuilds(unittest.TestCase):

    def setUp(self):
        self.jenkins = JenkinsJSONData()

    def test_get_downstream_builds(self):
        ret = jenkinsutils.get_downstream_builds(
            self.jenkins, 'http://10.0.0.1:8080/job/', 'my-job', 3)
        expected = [
            {'project': 'job-multiconfig',
             'number': 3,
             'url': 'http://10.0.0.1:8080/job/job-multiconfig/3/',
             'result': 'FAILURE', },
            {'project': 'job-freestyle',
             'number': 7,
             'url': 'http://10.0.0.1:8080/job/job-freestyle/7/',
             'result': 'SUCCESS', },
            {'project': 'freestyle-downstream',
             'number': 12,
             'url': 'http://10.0.0.1:8080/job/freestyle-downstream/12/',
             'result': 'UNSTABLE', },
        ]
        self.assertEqual(ret, expected)

    def test_get_downstream_builds_too_deep(self):
        ret = jenkinsutils.get_downstream_builds(self.jenkins,
                                                 'http://ip:8080/job/',
                                                 'my-job',
                                                 3, 11)
        self.assertEquals(ret, [])

    def test_filter_intermediate_job(self):
        blacklist = lambda name: ['job-freestyle']
        with patch('jlp.jenkinsutils.get_config_option',
                   blacklist):
            ret = jenkinsutils.get_downstream_builds(
                self.jenkins, 'http://10.0.0.1:8080/job/', 'my-job', 3)
        expected = [
            {'project': 'job-multiconfig',
             'number': 3,
             'url': 'http://10.0.0.1:8080/job/job-multiconfig/3/',
             'result': 'FAILURE', },
            {'project': 'freestyle-downstream',
             'number': 12,
             'url': 'http://10.0.0.1:8080/job/freestyle-downstream/12/',
             'result': 'UNSTABLE', },
        ]
        self.assertEqual(ret, expected)

    def test_filter_leaf_job(self):
        blacklist = lambda name: ['freestyle-downstream']
        with patch('jlp.jenkinsutils.get_config_option',
                   blacklist):
            ret = jenkinsutils.get_downstream_builds(
                self.jenkins, 'http://10.0.0.1:8080/job/', 'my-job', 3)
        expected = [
            {'project': 'job-multiconfig',
             'number': 3,
             'url': 'http://10.0.0.1:8080/job/job-multiconfig/3/',
             'result': 'FAILURE', },
            {'project': 'job-freestyle',
             'number': 7,
             'url': 'http://10.0.0.1:8080/job/job-freestyle/7/',
             'result': 'SUCCESS', },
        ]
        self.assertEqual(ret, expected)


class JenkinsUtilsIsMatrixJob(unittest.TestCase):

    def setUp(self):
        self.jenkins = JenkinsJSONData()

    def test_matrix_job(self):
        ret = jenkinsutils.is_matrix_job(
            self.jenkins,
            'http://10.0.0.1:8080/job/job-multiconfig/')
        self.assertEqual(ret, True)

    def test_matrix_job_with_build_number(self):
        ret = jenkinsutils.is_matrix_job(
            self.jenkins,
            'http://10.0.0.1:8080/job/job-multiconfig/32')
        self.assertEqual(ret, True)

    def test_non_matrix_job(self):
        ret = jenkinsutils.is_matrix_job(
            self.jenkins,
            'http://10.0.0.1:8080/job/job-freestyle/')
        self.assertEqual(ret, False)


class JenkinsUtilsGetExecutedBuilds(unittest.TestCase):
    def setUp(self):
        self.jenkins = JenkinsJSONData()
        self.build_url = self.jenkins.build_url

    def test_get_executed_builds_wrong_url(self):
        ret = jenkinsutils.get_executed_builds(self.jenkins,
                                               'wrong url')
        expected = None
        self.assertEqual(ret, expected)

    def test_get_executed_builds(self):
        ret = jenkinsutils.get_executed_builds(self.jenkins, self.build_url)
        expected = [
            {'output': 'http://10.0.0.1:8080/job/job-multiconfig/.' +
                       '/distribution=raring,flavor=amd64/3/console',
             'result': 'FAILURE', },
            {'output': 'http://10.0.0.1:8080/job/job-multiconfig/.' +
                       '/distribution=raring,flavor=i386/3/console',
             'result': 'FAILURE', },
            {'output': 'http://10.0.0.1:8080/job/job-freestyle/7/console',
             'result': u'SUCCESS', },
            {'output': 'http://10.0.0.1:8080/job/freestyle-downstream/' +
                       '12/console',
             'result': 'UNSTABLE'}
        ]
        self.assertEqual(ret, expected)


class JenkinsUtilsGetExecutedTestRuns(unittest.TestCase):
    def setUp(self):
        self.jenkins = JenkinsJSONData()

    def test_get_coverity_artifacts(self):
        ret = jenkinsutils.get_coverity_artifacts(
            self.jenkins,
            'http://10.0.0.1:8080/job/job-multiconfig/' +
            './distribution=raring,flavor=amd64/3/')
        expected = ['http://10.0.0.1:8080/job/job-multiconfig/./' +
                    'distribution=raring,flavor=amd64/3/artifact/' +
                    'results/coverity/CID_10895.html',
                    'http://10.0.0.1:8080/job/job-multiconfig/./' +
                    'distribution=raring,flavor=amd64/3/artifact/results' +
                    '/coverity/CID_10896.html']
        self.assertEqual(ret, expected)

    def test_get_deb_artifacts(self):
        ret = jenkinsutils.get_deb_artifacts(
            self.jenkins, 'http://10.0.0.1:8080/job/job-freestyle/7/')
        expected = 'http://10.0.0.1:8080/job/job-freestyle/7/artifact/' + \
                   'work/output/*zip*/output.zip'
        self.assertEqual(ret, expected)

    def test_no_deb_artifacts(self):
        ret = jenkinsutils.get_deb_artifacts(
            self.jenkins,
            'http://10.0.0.1:8080/job/freestyle-downstream/12/')
        expected = None
        self.assertEqual(ret, expected)


class JenkinsUtilsIsJobPublished(unittest.TestCase):
    def test_is_job_published_with_private_job(self):
        jenkins = MagicMock()
        jenkins.get_job_config = MagicMock(return_value='<b></b>')
        with patch('jlp.jenkinsutils.jenkins.Jenkins',
                   new=MagicMock(return_value=jenkins)):
            self.assertFalse(jenkinsutils.is_job_published('my-job'))

    def test_is_job_published_with_no_job(self):
        jenkins = MagicMock()
        jenkins.get_job_config = MagicMock(return_value='')
        with patch('jlp.jenkinsutils.jenkins.Jenkins',
                   new=MagicMock(return_value=jenkins)):
            self.assertFalse(jenkinsutils.is_job_published(None))

    def test_is_job_published_with_public_job(self):
        jenkins = MagicMock()
        jenkins.get_job_config = MagicMock(
            return_value='<hudson.plugins.build__publisher.BuildPublisher>' +
                         '</hudson.plugins.build__publisher.BuildPublisher>')
        with patch('jlp.jenkinsutils.jenkins.Jenkins',
                   new=MagicMock(return_value=jenkins)):
            self.assertTrue(jenkinsutils.is_job_published('my-job'))

    def test_is_job_published_with_exception(self):
        jenkins = MagicMock()
        jenkins.get_job_config.side_effect = Exception(
            'get_job_config exception')
        with patch('jlp.jenkinsutils.jenkins.Jenkins',
                   new=MagicMock(return_value=jenkins)):
            self.assertFalse(jenkinsutils.is_job_published('my-job'))


class JenkinsUtilsTestHideJenkinsUrl(unittest.TestCase):

    def test_hide_jenkins_url_public_url(self):
        public_jenkins_url = get_config_option('public_jenkins_url')
        urls = [
            ('http://10.0.0.1:8080/job/android-audiosystem/31/',
             public_jenkins_url + '/job/android-audiosystem/31/'),
            ('http://10.0.0.1:8080/job/android-ubuntu-matrix/4/',
             public_jenkins_url + '/job/android-ubuntu-matrix/4/'),
            ('https://10.0.0.1:8080/job/unity-ci/32/',
             public_jenkins_url + '/job/unity-ci/32/')
        ]

        is_job_published = lambda job: True
        for url, expected in urls:
            with patch('jlp.jenkinsutils.is_job_published',
                       new=is_job_published):
                self.assertEqual(jenkinsutils.hide_jenkins_url(url),
                                 expected)

    def test_hide_jenkins_url_hidden_url(self):
        urls = [
            'https://10.0.0.1:8080/job/android-audiosystem/31/',
            'http://10.0.0.1:8080/job/android-ubuntu-matrix/4/',
            'http://10.0.0.1:8080/job/unity-ci/32/',
        ]

        is_job_published = lambda job: False
        for url in urls:
            with patch('jlp.jenkinsutils.is_job_published',
                       new=is_job_published):
                self.assertEqual(jenkinsutils.hide_jenkins_url(url),
                                 url.replace('http://10.0.0.1:8080',
                                             'http://jenkins:8080'))

    def test_hide_jenkins_url_unknown_url(self):
        urls = [
            'http://12.0.0.2:8080/job/android-audiosystem/31/',
            'http://12.0.0.2:8080/job/android-ubuntu-matrix/4/',
            'https://12.0.0.2:8080/job/unity-ci/32/',
            'https://12.0.0.080/droid-audiosystem/31/',
            'http://12.0.0.080/droid-ubuntu-matrix/4/',
            'http://12.0.0.080/ity-ci/32/',
        ]

        is_job_published = lambda job: False
        for url in urls:
            with patch('jlp.jenkinsutils.is_job_published',
                       new=is_job_published):
                self.assertEqual(jenkinsutils.hide_jenkins_url(url),
                                 url)


class JenkinsUtilsGetParameters(unittest.TestCase):

    jenkins_params = {"property":
                      [{},
                       {"parameterDefinitions":
                        [{"defaultParameterValue":
                          {"name": "target_branch", "value": "lp:unity"},
                          "description": "lp target branch",
                          "name": "target_branch",
                          "type": "StringParameterDefinition"},
                         {"defaultParameterValue":
                          {"name": "packaging_branch", "value": ""},
                          "description": "lp packaging branch",
                          "name": "packaging_branch",
                          "type": "StringParameterDefinition"},
                         {"defaultParameterValue":
                          {"name": "ppa_target",
                           "value": "ppa:unity-team/staging"},
                          "description": "lp PPA",
                          "name": "ppa_target",
                          "type": "StringParameterDefinition"},
                         {"defaultParameterValue":
                          {"name": "distributions",
                           "value": "quantal,raring"},
                          "description": "Distributions to dput",
                          "name": "distributions",
                          "type": "StringParameterDefinition"},
                         {"defaultParameterValue":
                          {"name": "use_description_for_commit",
                           "value": True},
                          "description": "use description",
                          "name": "use_description_for_commit",
                          "type": "BooleanParameterDefinition"},
                         {"defaultParameterValue":
                          {"name": "version_string_format",
                           "value": "{UPSTREAM_VERSION}bzr{BZR_REVISION}"},
                          "description": "This is a template",
                          "name": "version_string_format",
                          "type": "StringParameterDefinition"},
                         {"defaultParameterValue":
                          {"name": "dput_hooks",
                           "value": "H10disable_powerpc"},
                          "description": "hack hooks used for dput",
                          "name": "dput_hooks",
                          "type": "StringParameterDefinition"}]}]}

    def setUp(self):
        self.jenkins = JenkinsJSONData()
        self.jenkins.get_json_data = MagicMock(
            return_value=self.jenkins_params)

    def test_get_parameter(self):
        ret = jenkinsutils.get_parameter(self.jenkins, 'http://ip/job/x',
                                         'target_branch')
        self.assertEqual(ret, 'lp:unity')

    def test_get_parameter_boolean(self):
        ret = jenkinsutils.get_parameter(self.jenkins, 'http://ip/job/x',
                                         'use_description_for_commit')
        self.assertTrue(ret)

    def test_get_parameter_default(self):
        ret = jenkinsutils.get_parameter(self.jenkins, 'http://ip/job/x',
                                         'missing_branch', 'lp:missing')
        self.assertEqual(ret, 'lp:missing')

    def test_get_parameter_key_error_exception(self):
        with self.assertRaises(KeyError) as err:
            jenkinsutils.get_parameter(self.jenkins, 'http://ip/job/x',
                                       'missing_branch')
        self.assertIn("jenkins job parameter [missing_branch]: not found",
                      "%s" % (err.exception))


class TestAddMergeParams(unittest.TestCase):
    def test_add_merge_params(self):
        """Verifies add_merge_params adds the needed fasttrack parameters"""
        params = {'landing_candidate': 'branch-name',
                  'merge_proposal': 'http://launchpad/mp',
                  'candidate_revision': 100}
        get_parameters = MagicMock()
        get_parameters.side_effect = ['lp:target', '', 'ppa:project/ppa',
                                      'quantal,raring', True, 'format',
                                      'H00hook']
        with patch('jlp.jenkinsutils.get_parameter', get_parameters):
            jenkinsutils.add_merge_params(None, None, params)
            self.assertEqual(params['target_branch'], 'lp:target')
            self.assertEqual(params['source_branch'], 'branch-name')
            self.assertEqual(params['revision'], 100)
            self.assertEqual(params['packaging_branch'], '')
            self.assertEqual(params['ppa_target'], 'ppa:project/ppa')
            self.assertEqual(params['distributions'], 'quantal,raring')
            self.assertEqual(params['buildjob_url'],
                             'Build was skipped, merge only')
            self.assertEqual(params['use_description_for_commit'], True)
            self.assertEqual(params['version_string_format'], 'format')
            self.assertEqual(params['hooks'], 'H00hook')


class TestStartJenkinsJob(unittest.TestCase):
    def setUp(self):
        self.launchpad_user = MagicMock()

    def test_start_jenkins_job(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('jlp.jenkinsutils.jenkins', new=jenkins):
            jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                           'jenkins_job', mp)
            self.assertEqual(j.build_job.call_count, 1)
            self.assertEqual(mp.nominateReviewer.call_count, 1)

    def test_start_jenkins_job_while_diff_is_not_yet_generated(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = None
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)

        with patch('jlp.jenkinsutils.jenkins', new=jenkins):
            ret = jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                                 'jenkins_job', mp)
            self.assertFalse(ret)
            self.assertEqual(j.build_job.call_count, 0)

    def test_start_jenkins_job_and_raise_exception(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        j.build_job.side_effect = JenkinsException("jenkinsException")
        jenkinsInstance = MagicMock()
        jenkinsInstance.Jenkins = MagicMock(return_value=j)
        jenkinsInstance.JenkinsException = JenkinsException
        with patch('jlp.jenkinsutils.jenkins', new=jenkinsInstance):
            jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                           'jenkins_job', mp)
            self.assertEqual(mp.createComment.call_count, 1)

    def test_start_jenkins_job_and_raise_unauthorized(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.nominateReviewer.side_effect = Unauthorized("unauthorized", '')
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkinsInstance = MagicMock()
        jenkinsInstance.Jenkins = MagicMock(return_value=j)
        with patch('jlp.jenkinsutils.jenkins', new=jenkinsInstance):
            ret = jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                                 'jenkins_job', mp)
            self.assertEqual(mp.createComment.call_count, 0)
            self.assertEqual(ret, False)

    def test_start_nonbuildable_jenkins_job(self):
        mp = MagicMock()
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': False})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('jlp.jenkinsutils.jenkins', new=jenkins):
            self.assertEqual(
                jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                               'jenkins_job', mp),
                False)
            self.assertEqual(j.build_job.call_count, 0)
            self.assertEqual(mp.nominateReviewer.call_count, 0)
            self.assertEqual(mp.createComment.call_count, 0)

    @patch('jlp.jenkinsutils.add_merge_params')
    def test_start_jenkins_job_with_fasttrack_merge(self, add_merge_params):
        """Verifies that the merge job is called with merge only input"""
        json_jenkins = MagicMock()
        json_jenkins.get_json_data = MagicMock(
            return_value={'downstreamProjects':
                          [{'name': 'generic-cleanup'},
                           {'name': 'generic-land'}]})
        get_json_jenkins = MagicMock(return_value=json_jenkins)
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('jlp.jenkinsutils.jenkins', new=jenkins), \
                patch('jlp.jenkinsutils.get_json_jenkins',
                      new=get_json_jenkins):
            jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                           'jenkins_job', mp, True)
            self.assertEqual(j.build_job.call_count, 1)
            self.assertEqual(j.build_job.call_args[0][0], 'generic-land')

    @patch('jlp.launchpadutils.is_merge_only', new=lambda x: False)
    def test_start_jenkins_job_with_fasttrack_merge_needs_build(self):
        """Verifies that the jenkins job is called when build needed"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        j = MagicMock()
        j.get_job_info = MagicMock(return_value={'buildable': True})
        jenkins = MagicMock()
        jenkins.Jenkins = MagicMock(return_value=j)
        with patch('jlp.jenkinsutils.jenkins', new=jenkins):
            jenkinsutils.start_jenkins_job(self.launchpad_user, 'url',
                                           'jenkins_job', mp, True)
            self.assertEqual(j.build_job.call_count, 1)
            self.assertEqual(j.build_job.call_args[0][0], 'jenkins_job')


class TestGetResultLine(TestWithScenarios):
    scenarios = [
        ('success',
         {'result': 'SUCCESS',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected': "\n    SUCCESS: http://10.0.0.1:8080/job/my-job/13"}),
        ('success_with_slash',
         {'result': 'SUCCESS',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console/',
          'expected': "\n    SUCCESS: http://10.0.0.1:8080/job/my-job/13"}),
        ('unstable',
         {'result': 'UNSTABLE',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected': "\n    UNSTABLE: http://10.0.0.1:8080/job/my-job/13"}),
        ('failed',
         {'result': 'FAILURE',
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected':
          "\n    FAILURE: http://10.0.0.1:8080/job/my-job/13/console"}),
        ('none',
         {'result': None,
          'url': 'http://10.0.0.1:8080/job/my-job/13/console',
          'expected':
          "\n    None: http://10.0.0.1:8080/job/my-job/13/console"}),
    ]

    def test_get_result_line(self):
        result = jenkinsutils._get_result_line(self.result, self.url)
        self.assertEqual(result, self.expected)


class LaunchpadTriggerGetExecutedTestRuns(unittest.TestCase):

    executed_test_runs_message = dedent(
        """
     Executed test runs:
         FAILURE: http://{hostname}/job/job-multiconfig/./""" +
        """distribution=raring,flavor=amd64/3/console
         FAILURE: http://{hostname}/job/job-multiconfig/./""" +
        """distribution=raring,flavor=i386/3/console
         SUCCESS: http://{hostname}/job/job-freestyle/7
             deb: http://{hostname}/job/job-freestyle/7/artifact/work/""" +
        """output/*zip*/output.zip
         UNSTABLE: http://{hostname}/job/freestyle-downstream/12
     Coverity artifacts:
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=amd64/3/artifact/results/coverity/CID_10895.html
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=amd64/3/artifact/results/coverity/CID_10896.html
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=i386/3/artifact/results/coverity/CID_10895.html
         http://{hostname}/job/job-multiconfig/./distribution=""" +
        """raring,flavor=i386/3/artifact/results/coverity/""" +
        """CID_10896.html""")

    def _get_executed_test_runs_message(self, expected_jenkins_hostname):
        return self.executed_test_runs_message.format(
            hostname=expected_jenkins_hostname)

    def setUp(self):
        super(LaunchpadTriggerGetExecutedTestRuns, self).setUp()
        jenkins = JenkinsJSONData()
        self.build_url = jenkins.build_url
        self.get_json_jenkins_patch = patch(
            'jlp.jenkinsutils.get_json_jenkins', new=lambda: jenkins)
        self.get_json_jenkins_patch.start()

    def tearDown(self):
        self.get_json_jenkins_patch.stop()

    def test_get_executed_test_runs_message(self):
        ret = jenkinsutils.get_executed_test_runs_message(self.build_url)
        self.assertEqual(ret,
                         self._get_executed_test_runs_message('jenkins:8080'))

    def test_format_message_for_mp_update(self):
        test_runs = self._get_executed_test_runs_message('jenkins:8080')
        expected = dedent("""\
     mymessage
     http://jenkins:8080/job/my-job/3/{test_runs}

     Click here to trigger a rebuild:
     http://jenkins:8080/job/my-job/3/rebuild
             """).format(test_runs=test_runs)
        actual = jenkinsutils.format_message_for_mp_update(
            self.build_url,
            "mymessage\n")
        self.assertEquals(actual, expected)

    def test_format_message_with_empty_message(self):
        test_runs = self._get_executed_test_runs_message('jenkins:8080')
        expected = dedent("""\
     http://jenkins:8080/job/my-job/3/{test_runs}

     Click here to trigger a rebuild:
     http://jenkins:8080/job/my-job/3/rebuild
              """).format(test_runs=test_runs)

        actual = jenkinsutils.format_message_for_mp_update(
            self.build_url)
        self.assertEquals(actual, expected)

    def test_format_message_with_publishing_job(self):
        test_runs = self._get_executed_test_runs_message(
            'jenkins.qa.ubuntu.com')
        expected = dedent("""\
     http://jenkins.qa.ubuntu.com/job/my-job/3/{test_runs}

     Click here to trigger a rebuild:
     http://jenkins:8080/job/my-job/3/rebuild
              """).format(test_runs=test_runs)

        with patch("jlp.jenkinsutils.is_job_published",
                   new=lambda x: True):
            actual = jenkinsutils.format_message_for_mp_update(
                self.build_url)
            self.assertEquals(actual, expected)


class TestTriggerCi(unittest.TestCase):
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs',
           new=lambda x: False)
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_ci_with_wrong_user(self, start_jenkins_job):
        """If user is not allowed to trigger jobs than a jenkins job must not
        be started."""
        lp_handle = MagicMock()
        jenkinsutils.trigger_ci_build(lp_handle, [MagicMock()], 'job', 'url')
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_ci_with_no_mps(self, start_jenkins_job):
        """If there are no merge proposals then no jenkins job must be
        started"""
        lp_handle = MagicMock()
        jenkinsutils.trigger_ci_build(lp_handle, [], 'job', 'url')
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.launchpadutils.testing_in_progress', new=lambda x: True)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs')
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_ci_while_testing_in_progress(self,
                                                  start_jenkins_job,
                                                  mock):
        """If testing is already in progress then no job must be started"""
        lp_handle = MagicMock()
        jenkinsutils.trigger_ci_build(lp_handle, [MagicMock()], 'job', 'url')
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.launchpadutils.latest_candidate_validated',
           new=lambda x, y: True)
    @patch('jlp.launchpadutils.testing_in_progress', new=lambda x: False)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs')
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_ci_while_Revision_validated(self,
                                                 start_jenkins_job,
                                                 mock):
        """If the latest revision was already validated then no jenkins job
        must be started"""
        lp_handle = MagicMock()
        jenkinsutils.trigger_ci_build(lp_handle, [MagicMock()], 'job', 'url')
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.launchpadutils.latest_candidate_validated',
           new=lambda x, y: False)
    @patch('jlp.launchpadutils.testing_in_progress', new=lambda x: False)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs')
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_ci_while_Revision_not_validated(self,
                                                     start_jenkins_job,
                                                     mock):
        """If testing it not in progress, user is allowed to trigger jobs and
        the latest revision of MP is not yet validated then a jenkins job must
        be started
        """
        lp_handle = MagicMock()
        jenkinsutils.trigger_ci_build(lp_handle, [MagicMock()], 'job', 'url')
        self.assertEqual(start_jenkins_job.call_count, 1)


class TestTriggerAl(unittest.TestCase):
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs',
           new=lambda x: False)
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_al_without_permissions(self, start_jenkins_job):
        """If user it not allowed to start jobs then no jenkins job must be
        started"""
        lp_handle = MagicMock()
        jenkinsutils.trigger_al_build(lp_handle,
                                      [MagicMock(), MagicMock(), MagicMock()],
                                      'job', 'url', None)
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.launchpadutils.is_commit_message_set',
           new=lambda x, y, z: False)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs',
           new=lambda x: True)
    @patch('jlp.launchpadutils.change_mp_status')
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_al_with_no_commit_message(self, start_jenkins_job,
                                               change_mp_status):
        """If no commit message is set no jenkins job must be started and
        status of merge proposal must be changed to NEEDS_FIXING"""
        jenkinsutils.trigger_al_build(MagicMock(),
                                      [MagicMock(), MagicMock(), MagicMock()],
                                      'job', 'url', None)
        self.assertEqual(start_jenkins_job.call_count, 0)
        # the new vote is the 6th argument in change_mp_status (and we count
        # from 0)
        new_vote_position = 5
        self.assertEqual(change_mp_status.call_args[0][new_vote_position],
                         launchpadutils.LaunchpadVote.NEEDS_FIXING)

    @patch('jlp.launchpadutils.is_commit_message_set',
           new=lambda x, y, z: True)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs',
           new=lambda x: True)
    @patch('jlp.launchpadutils.testing_in_progress',
           new=lambda mp: True)
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_al_while_testing_in_progress(self, start_jenkins_job):
        """When testing is in progress jenkins job must not be started"""
        jenkinsutils.trigger_al_build(MagicMock(),
                                      [MagicMock(), MagicMock(), MagicMock()],
                                      'job', 'url', None)
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.launchpadutils.get_json_jenkins')
    @patch('jlp.launchpadutils.unapproved_prerequisite_exists',
           new=lambda x: True)
    @patch('jlp.launchpadutils.testing_in_progress',
           new=lambda x: False)
    @patch('jlp.launchpadutils.is_commit_message_set',
           new=lambda x, y, z: True)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs',
           new=lambda x: True)
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_al_with_unapproved_prerequisite(self,
                                                     start_jenkins_job,
                                                     json_jenkins):
        """If unapproved prerequisite exists the jenkins job must not be
        started"""
        jenkinsutils.trigger_al_build(MagicMock(),
                                      [MagicMock(), MagicMock(), MagicMock()],
                                      'job', 'url', None)
        self.assertEqual(start_jenkins_job.call_count, 0)

    @patch('jlp.launchpadutils.is_commit_message_set',
           new=lambda x, y, z: True)
    @patch('jlp.launchpadutils.unapproved_prerequisite_exists',
           new=lambda x: False)
    @patch('jlp.launchpadutils.testing_in_progress',
           new=lambda x: False)
    @patch('jlp.launchpadutils.user_allowed_to_trigger_jobs',
           new=lambda x: True)
    @patch('jlp.jenkinsutils.start_jenkins_job')
    def test_trigger_al(self, start_jenkins_job):
        """If commit message is set, there is no unapproved prerequeisite, no
        testing is in progress, user is alloswed to trigger jobs then a
        jenkins job must be started"""
        jenkinsutils.trigger_al_build(MagicMock(),
                                      [MagicMock(), MagicMock(), MagicMock()],
                                      'job', 'url', None)
        self.assertEqual(start_jenkins_job.call_count, 3)


class LaunchpadTriggerAutolandingWithTrustedReviewer(unittest.TestCase):
    def setUp(self):
        self.user_allowed_to_trigger_jobs_patch = patch(
            'jlp.launchpadutils.user_allowed_to_trigger_jobs',
            new=lambda x: x)
        self.user_allowed_to_trigger_jobs_patch.start()
        self.is_commit_message_set_patch = patch(
            'jlp.launchpadutils.is_commit_message_set',
            new=lambda x, y, z: True)
        self.is_commit_message_set = self.is_commit_message_set_patch.start()
        self.start_jenkins_job_patch = patch(
            'jlp.jenkinsutils.start_jenkins_job')
        self.start_jenkins_job_patch.start()
        #return True so we stop the flow; We just want to test the permissions
        #above and not the whole trigger_al_build method
        test_in_progress = MagicMock(return_value=True)
        self.testing_in_progress_patch = patch(
            'jlp.launchpadutils.testing_in_progress',
            new=test_in_progress)
        self.testing_in_progress = self.testing_in_progress_patch.start()

    def tearDown(self):
        self.is_commit_message_set_patch.stop()
        self.testing_in_progress_patch.stop()
        self.user_allowed_to_trigger_jobs_patch.stop()
        self.start_jenkins_job_patch.stop()

    def test_trigger_al_with_trusted_reviewer(self):
        mp = MagicMock()
        #user_allowed_to_trigger_jobs now returns what it gets as an argument
        #(see the lambda in constructor)
        #We want to return False for registrant (i.e. person who create MP is
        #not allowed to trigger jobs in our infrastructure)
        #We want to return True for reviewer (i.e. the person who reviewed is
        #trusted and we can run the code on his behalf)
        mp.registrant = False
        mp.reviewer = True
        jenkinsutils.trigger_al_build(MagicMock(), [mp], 'job', 'url', None)
        self.assertEqual(self.testing_in_progress.call_count, 1)

    def test_trigger_al_with_trusted_registrant(self):
        mp = MagicMock()
        #user_allowed_to_trigger_jobs now returns what it gets as an argument
        #(see the lambda in constructor)
        #in this test-case even if the reviewer is not trusted we trust the
        #registrant so we are good to go
        mp.registrant = True
        mp.reviewer = False
        jenkinsutils.trigger_al_build(MagicMock(), [mp], 'job', 'url', None)
        self.assertEqual(self.testing_in_progress.call_count, 1)

    def test_trigger_al_with_untrusted_people(self):
        mp = MagicMock()
        #user_allowed_to_trigger_jobs now returns what it gets as an argument
        #(see the lambda in constructor)
        #in this case none of the persons are trusted and therefore
        #we should stop and testing_in_progress shouldn't be executed at all
        mp.registrant = False
        mp.reviewer = False
        jenkinsutils.trigger_al_build(MagicMock(), [mp], 'job', 'url', None)
        self.assertEqual(self.testing_in_progress.call_count, 0)
