import unittest
from jlp import launchpadutils, MergeProposalReview, get_config_option
from mock import MagicMock, patch
from testscenarios import TestWithScenarios
from textwrap import dedent


class LaunchpadUtilsTests(unittest.TestCase):
    launchpadAgent = None

    def test_get_branch_from_mp(self):
        branch = launchpadutils.get_branch_from_mp(
            'https://copde.launchpad.net/~user/project/name/+merge/12345')
        expected = 'https://copde.launchpad.net/~user/project/name/'
        self.assertEqual(branch, expected)

    def test_get_branch_from_mp_invalid_url(self):
        branch = launchpadutils.get_branch_from_mp(
            'https://copde.launchpad.net/~user/project/name/merge/12345')
        self.assertEqual(branch, None)

    def test_change_mp_status_with_no_vote(self):
        mp = MagicMock()
        launchpadutils.change_mp_status(mp, 'Status', 'Message',
                                        'Subject', revision=1)
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            subject='Subject',
            content='Message')
        mp.setStatus.assert_called_once_with(status='Status', revid=1)

    def test_change_mp_status_with_approve(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        launchpadutils.change_mp_status(mp, 'Status', 'Message',
                                        'Subject', revision=1,
                                        vote='Approve')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            subject='Subject',
            content='Message', vote='Approve')
        mp.setStatus.assert_called_once_with(status='Status', revid=1)

    def test_change_mp_status_with_disapprove(self):
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        launchpadutils.change_mp_status(mp, 'Status', 'Message',
                                        'Subject', revision=1,
                                        vote='Disapprove')
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            subject='Subject',
            content='Message', vote='Disapprove')
        mp.setStatus.assert_called_once_with(status='Status', revid=1)

    def test_close_bugs(self):
        mp = MagicMock()
        mp.target_branch.project.web_link = 'http://launchpad/jenkins'
        mp.target_branch.project.name = 'project'
        mp.source_branch.linked_bugs_collection = [MagicMock(), MagicMock()]
        mp.target_branch.display_name = 'lp:project'
        mp.merged_revno = '12'
        mp.target_branch.self_link = 'http://project/project'
        milestone = MagicMock()
        milestone.series_target.branch_link = mp.target_branch.self_link
        milestone.name = '5.0'
        mp.target_branch.project.active_milestones_collection = [milestone]
        for b in mp.source_branch.linked_bugs_collection:
            b.bug_tasks_collection = [MagicMock(), MagicMock(), MagicMock()]
            bug_task = b.bug_tasks_collection[1]
            bug_task.target.web_link = mp.target_branch.project.web_link

        launchpadutils.close_bugs(mp)
        for b in mp.source_branch.linked_bugs_collection:
            b.bug_tasks_collection[1].lp_save.assert_called_once_with()
            b.newMessage.assert_called_once_with(
                content='Fix committed into lp:project at revision ' +
                '12, scheduled for release in project, milestone 5.0')

    def test_get_reviews_no_reviews(self):
        mp = MagicMock()
        mp.votes = []
        self.assertTrue(launchpadutils.get_reviews(mp) is None)

    def test_get_reviews(self):
        mp = MagicMock()
        mp.votes = [MagicMock(), MagicMock(), MagicMock()]
        for vote in mp.votes:
            vote.reviewer.display_name = 'jenkins'
            vote.comment.vote = 'Approve'
        reviews = launchpadutils.get_reviews(mp)
        for review in reviews:
            self.assertEquals(review, 'jenkins;Approve')

    def test_unapproved_prerequisite_no_prerequisite(self):
        mp = MagicMock()
        mp.prerequisite_branch = None
        self.assertFalse(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_too_many_proposals(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock(), MagicMock(), MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Approved'
        mp.prerequisite_branch = prereq
        self.assertTrue(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_no_proposal_for_prerequisite(self):
        mp = MagicMock()
        prereq = MagicMock()
        prereq.landing_targets = []
        mp.prerequisite_branch = prereq
        self.assertTrue(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_not_merged_yet(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Approved'
        mp.prerequisite_branch = prereq
        self.assertTrue(launchpadutils.unapproved_prerequisite_exists(mp))

    def test_unapproved_prerequisite_merged(self):
        mp = MagicMock()
        mp.target_branch = MagicMock()
        mp.target_branch.web_link = 'url'
        prereq = MagicMock()
        prereq.landing_targets = [MagicMock()]
        for t in prereq.landing_targets:
                t.target_branch.web_link = 'url'
                t.queue_status = 'Merged'
        mp.prerequisite_branch = prereq
        self.assertFalse(launchpadutils.unapproved_prerequisite_exists(mp))


class TestUserAllowedToTriggerJobs(unittest.TestCase):
    def test_user_not_allowed_to_trigger_job(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        membership = MagicMock()
        membership.team.name = 'random-team'
        registrant.memberships_details = [membership]
        result = launchpadutils.user_allowed_to_trigger_jobs(registrant)
        self.assertFalse(result)

    def test_user_member_of_allowed_team(self):
        registrant = MagicMock()
        registrant.name = 'random-user'
        membership = MagicMock()
        membership.team.name = 'ps-jenkins'
        registrant.memberships_details = [membership]
        result = launchpadutils.user_allowed_to_trigger_jobs(registrant)
        self.assertTrue(result)

    def test_user_allowed_to_trigger(self):
        registrant = MagicMock()
        registrant.name = 'ps-jenkins'
        membership = MagicMock()
        membership.team.name = ''
        registrant.memberships_details = [membership]
        result = launchpadutils.user_allowed_to_trigger_jobs(registrant)
        self.assertTrue(result)


class TestIsMergeOnly(unittest.TestCase):
    def test_is_merge_only(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertTrue(launchpadutils.is_merge_only(mp))

    def test_is_merge_only_not_changelog(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/rules': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertFalse(launchpadutils.is_merge_only(mp))

    def test_is_merge_only_multiple_files(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        diff_stat = {u'debian/changelog': 'diff text',
                     u'debian/rules': 'diff text'}
        mp.preview_diff.diffstat = diff_stat
        self.assertFalse(launchpadutils.is_merge_only(mp))

    def test_is_merge_only_no_preview_diff(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = None
        with self.assertRaises(LookupError):
            launchpadutils.is_merge_only(mp)

    def test_is_merge_only_no_diffstat(self):
        """Verifies that the a changelog only change is merge only"""
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        mp.preview_diff = MagicMock()
        mp.preview_diff.diffstat = None
        with self.assertRaises(LookupError):
            launchpadutils.is_merge_only(mp)


class TestGetCommitMessage(TestWithScenarios):
    scenarios = [
        ('no_commit_message',
         {'commit_message': None,
          'description': 'description',
          'use_description': False,
          'expected': None}),
        ('no_commit_message_fallback_to_description',
         {'commit_message': None,
          'description': 'description',
          'use_description': True,
          'expected': 'description'}),
        ('commit_message_and_description',
         {'commit_message': 'commit_message',
          'description': 'description',
          'use_description': True,
          'expected': 'commit_message'}),
        ('no_commit_message_and_no_description',
         {'commit_message': '',
          'description': '',
          'use_description': True,
          'expected': None})]

    def test_get_commit_message(self):
        mp = MagicMock()
        mp.commit_message = self.commit_message
        mp.description = self.description
        use_description = self.use_description

        message = launchpadutils.get_commit_message(mp, use_description)
        self.assertEqual(message, self.expected)


class LaunchpadTriggerIsCommitMessageSet(unittest.TestCase):
    def _setup_json_data(self, json_data):
        jenkins = MagicMock()
        jenkins.get_json_data = MagicMock(return_value=json_data)
        self.get_json_jenkins_patch = patch(
            'jlp.launchpadutils.get_json_jenkins',
            new=MagicMock(return_value=jenkins))
        self.get_json_jenkins_mock = self.get_json_jenkins_patch.start()

    def _setup_get_commit_message(self, return_value):
        get_commit_message = MagicMock(return_value=return_value)
        self.get_commit_message_patch = patch(
            'jlp.launchpadutils.get_commit_message',
            new=get_commit_message)
        self.get_commit_message_mock = self.get_commit_message_patch.start()

    def _cleanup(self):
        self.get_commit_message_patch.stop()
        self.get_json_jenkins_patch.stop()

    def test_use_description_but_still_no_commit_message(self):
        json_data = {'property': ['dummy', {'parameterDefinitions': [{
            'name': 'use_description_for_commit',
            'defaultParameterValue': {'value': True}}]}]}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(False)
        ret = launchpadutils.is_commit_message_set(None, 'job', 'url')
        self.get_commit_message_mock.assert_called_once_with(None, True)
        self.assertFalse(ret)
        self._cleanup()

    def test_no_use_description_but_commit_message_set(self):
        json_data = {'property': ['dummy', {'parameterDefinitions': [{
            'name': 'use_description_for_commit',
            'defaultParameterValue': {'value': False}}]}]}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(True)
        ret = launchpadutils.is_commit_message_set(None, 'job', 'url')
        self.get_commit_message_mock.assert_called_once_with(None, False)
        self.assertTrue(ret)
        self._cleanup()

    def test_use_description_flag_not_available(self):
        json_data = {}
        self._setup_json_data(json_data)
        self._setup_get_commit_message(True)
        ret = launchpadutils.is_commit_message_set(None, 'job', 'url')
        self.get_commit_message_mock.assert_called_once_with(None, False)
        self.assertTrue(ret)
        self._cleanup()

    def test_testing_not_in_progress_approved(self):
        """ If there is no MergeProposalReview lock then testing_in_progress
        must return False.
        """
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        review = MergeProposalReview(mp)
        review.unlock()
        self.assertFalse(launchpadutils.testing_in_progress(mp))

    def test_testing_not_in_progress_abstain(self):
        """ If a MergeProposalReview lock exists then testing_in_progress
        must return True and it must report the correct jenkins job which
        holds the lock.
        """
        mp = MagicMock()
        mp.web_link = 'http://my-example-url.com'
        review = MergeProposalReview(mp)
        review.lock('my-job')
        self.assertTrue(launchpadutils.testing_in_progress(mp))
        self.assertEquals(review.get_jenkins_job(), 'my-job')


class TestLatestReview(unittest.TestCase):
    def setUp(self):
        self.launchpad_user = MagicMock()
        self.jenkinsUserName = 'name'
        self.launchpad_user.name = self.jenkinsUserName

    def test_get_latest_review_no_comments(self):
        mp = MagicMock()
        mp.all_comments = []
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 0)

    def test_get_latest_review_unrelated_comment(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = 'unrelated comment'
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 0)

    def test_get_latest_review_which_passed(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = "PASSED: Continuous integration, rev:47\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 47)

    def test_get_latest_review_which_failed(self):
        mp = MagicMock()
        comment = MagicMock()
        comment.author.name = self.jenkinsUserName
        comment.vote_tag = get_config_option('launchpad_review_type')
        comment.message_body = "FAILED: Continuous integration, rev:47\n" + \
                               "URL:..."
        mp.all_comments = [comment]
        self.assertEqual(
            launchpadutils.get_latest_review(self.launchpad_user, mp), 47)

    @patch('jlp.launchpadutils.get_latest_review', new=lambda x, y: 60)
    def test_new_mp_candidate_revision_in_queue(self):
            #60 is the revision which was validated as last
            mp = MagicMock()
            #this is the current branch revision
            mp.source_branch.revision_count = 61
            #latest_candidate_validated should return False as there is a new
            #revision in the queue (61)
            ret = launchpadutils.latest_candidate_validated(
                self.launchpad_user, mp)
            self.assertFalse(ret)

    @patch('jlp.launchpadutils.get_latest_review', new=lambda x, y: 60)
    def test_mp_already_validated(self):
            #60 is the revision which was validated as last
            mp = MagicMock()
            #this is the current branch revision
            mp.source_branch.revision_count = 60
            #latest_candidate_validated should return True. This rev was
            #already validated
            ret = launchpadutils.latest_candidate_validated(
                self.launchpad_user, mp)
            self.assertTrue(ret)


class TestApproveAndDisapprove(unittest.TestCase):
    jenkinsUserName = 'jenkins-user'

    def setUp(self):
        user = MagicMock()
        user.name = self.jenkinsUserName
        self.get_executed_test_runs_message_patch = patch(
            'jlp.jenkinsutils.get_executed_test_runs_message',
            new=lambda x: '')
        self.get_executed_test_runs_message_patch.start()

    def tearDown(self):
        self.get_executed_test_runs_message_patch.stop()

    def test_approve_mp(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        launchpadutils.approve_mp(mp, '1', 'http://url')
        subject = 'PASSED: Continuous integration, rev:1'
        content = dedent('''\
                            {subject}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(subject=subject)
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.APPROVE,
            subject=subject,
            content=content)

    def test_disapprove_mp(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        launchpadutils.disapprove_mp(mp, '1', 'http://url')
        subject = 'FAILED: Continuous integration, rev:1'
        content = dedent('''\
                            {subject}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(subject=subject)
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)

    def test_disapprove_mp_with_reason(self):
        mp = MagicMock()
        mp.web_link = 'http://mp'
        subject = 'FAILED: Continuous integration, rev:1'
        reason = "No commit message set\nDo something\n"
        content = dedent('''\
                            {subject}
                            {reason}
                            http://url

                            Click here to trigger a rebuild:
                            http://url/rebuild
                            ''').format(subject=subject, reason=reason)

        launchpadutils.disapprove_mp(mp,
                                     '1',
                                     'http://url',
                                     reason=reason)
        mp.createComment.assert_called_once_with(
            review_type=get_config_option('launchpad_review_type'),
            vote=launchpadutils.LaunchpadVote.NEEDS_FIXING,
            subject=subject,
            content=content)


class TestGetMergeProposalsWithInvalidBranch(unittest.TestCase):
    branch_with_mps = None

    def setUp(self):
        branch = MagicMock()
        branch.landing_candidates = []
        for i in range(1, 10):
            mp = MagicMock()
            mp.queue_status = i
            branch.landing_candidates.append(mp)
        self.branch_with_mps = branch

    def _get_launchpad(self, getByUrl_result=None):
        launchpad = MagicMock()
        launchpad.branches.getByUrl = lambda url: getByUrl_result
        return launchpad

    def test_get_merge_proposals_with_invalid_branch(self):
        launchpad = self._get_launchpad(None)
        ret = launchpadutils.get_merge_proposals(launchpad, '', '')
        self.assertEqual(ret, [])

    def test_get_merge_proposals_odd(self):
        launchpad = self._get_launchpad(self.branch_with_mps)
        mps = launchpadutils.get_merge_proposals(launchpad, '', [1, 3, 5, 7])
        self.assertEqual(len(mps), 4)

    def test_get_merge_proposal_two(self):
        launchpad = self._get_launchpad(self.branch_with_mps)
        mps = launchpadutils.get_merge_proposals(launchpad, '', [2])
        self.assertEqual(len(mps), 1)
        self.assertEqual(mps[0].queue_status, 2)


class TestGetMpHandleFromURL(unittest.TestCase):
    @patch("jlp.launchpadutils.get_branch_from_mp", new=lambda x: None)
    def test_unable_to_get_branch_link(self):
        """If there is no way how to get branch link from the merge proposal
        url then get_mp_handle_from_url must return None"""
        ret = launchpadutils.get_mp_handle_from_url(MagicMock(), 'http://')
        self.assertEquals(ret, None)

    @patch("jlp.launchpadutils.get_branch_from_mp", new=lambda x: 'http://')
    def test_unable_to_get_branch_handle(self):
        """getByURL launchpad call returns None then get_mp_handle_from_url
        must return None too"""
        launchpad = MagicMock()
        launchpad.branches.getByUrl = lambda url: None
        ret = launchpadutils.get_mp_handle_from_url(launchpad, 'http://')
        self.assertEquals(ret, None)

    @patch("jlp.launchpadutils.get_branch_from_mp", new=lambda x: 'http://')
    def test_get_mp_handle_from_url(self):
        """If there is a merge proposal for the given merge proposal URL then
        get_mp_handle_from_url must return the respective merge proposal
        handle"""

        mp_url = 'http://my-url'
        launchpad = MagicMock()
        branch = MagicMock
        mp = MagicMock()
        mp.web_link = mp_url
        branch.landing_targets = [mp]
        launchpad.branches.getByUrl = lambda url: branch
        ret = launchpadutils.get_mp_handle_from_url(launchpad, mp_url)
        self.assertEquals(ret, mp)
